import sys
import warnings

if not sys.warnoptions:
    warnings.simplefilter('ignore')

import tensorflow as tf
import numpy as np
from ..texts._text_functions import (
    str_idx,
    entities_textcleaning,
    char_str_idx,
    generate_char_seq,
    language_detection_textcleaning,
)
from ..stem import _classification_textcleaning_stemmer


def _convert_sparse_matrix_to_sparse_tensor(X, got_limit = True, limit = 5):
    coo = X.tocoo()
    indices = np.mat([coo.row, coo.col]).transpose()
    if got_limit:
        coo.data[coo.data > limit] = limit
    return (
        tf.SparseTensorValue(indices, coo.col, coo.shape),
        tf.SparseTensorValue(indices, coo.data, coo.shape),
    )


class _LANG_MODEL:
    def __init__(self):
        self.X = tf.sparse_placeholder(tf.int32)
        self.W = tf.sparse_placeholder(tf.int32)
        self.Y = tf.placeholder(tf.int32, [None])
        embeddings = tf.Variable(tf.truncated_normal([660726, 40]))
        embed = tf.nn.embedding_lookup_sparse(
            embeddings, self.X, self.W, combiner = 'mean'
        )
        self.logits = tf.layers.dense(embed, 4)


class _SPARSE_SOFTMAX_MODEL:
    def __init__(self, output_size, embedded_size, vocab_size):
        self.X = tf.sparse_placeholder(tf.int32)
        self.W = tf.sparse_placeholder(tf.int32)
        self.Y = tf.placeholder(tf.int32, [None])
        embeddings = tf.Variable(
            tf.truncated_normal([vocab_size, embedded_size])
        )
        embed = tf.nn.embedding_lookup_sparse(
            embeddings, self.X, self.W, combiner = 'mean'
        )
        self.logits = tf.layers.dense(embed, output_size)


class DEPENDENCY:
    def __init__(
        self,
        X,
        X_char,
        logits,
        logits_depends,
        settings,
        sess,
        model,
        transitions,
        transitions_depends,
        features,
    ):
        self._X = X
        self._X_char = X_char
        self._logits = logits
        self._logits_depends = logits_depends
        self._settings = settings
        self._sess = sess
        self._model = model
        self._settings['idx2tag'] = {
            int(k): v for k, v in self._settings['idx2tag'].items()
        }
        self.transitions, self.transitions_depends, self.features = self._sess.run(
            [transitions, transitions_depends, features]
        )

    def print_transitions_tag(self, top_k = 10):
        """
        Print important top-k transitions for tagging dependency

        Parameters
        ----------
        top_k : int
        """
        if not isinstance(top_k, int):
            raise ValueError('input must be an integer')
        print('Top-%d likely transitions:' % (top_k))
        indices = np.argsort(self.transitions.flatten())[::-1]
        top_trans = [
            np.unravel_index(i, self.transitions.shape) for i in indices[:top_k]
        ]
        for i in range(top_k):
            print(
                '%s -> %s: %f'
                % (
                    self._settings['idx2tag'][top_trans[i][0]],
                    self._settings['idx2tag'][top_trans[i][1]],
                    self.transitions[top_trans[i]],
                )
            )

        bottom_trans = [
            np.unravel_index(i, self.transitions.shape)
            for i in indices[::-1][:top_k]
        ]
        print('\nTop-%d unlikely transitions:' % (top_k))
        for i in range(top_k):
            print(
                '%s -> %s: %f'
                % (
                    self._settings['idx2tag'][bottom_trans[i][0]],
                    self._settings['idx2tag'][bottom_trans[i][1]],
                    self.transitions[bottom_trans[i]],
                )
            )

    def print_transitions_index(self, top_k = 10):
        """
        Print important top-k transitions for indexing dependency

        Parameters
        ----------
        top_k : int
        """
        if not isinstance(top_k, int):
            raise ValueError('input must be an integer')
        print('Top-%d likely transitions:' % (top_k))
        indices = np.argsort(self.transitions_depends.flatten())[::-1]
        top_trans = [
            np.unravel_index(i, self.transitions_depends.shape)
            for i in indices[:top_k]
        ]
        for i in range(top_k):
            print(
                '%d -> %d: %f'
                % (
                    top_trans[i][0],
                    top_trans[i][1],
                    self.transitions_depends[top_trans[i]],
                )
            )

        bottom_trans = [
            np.unravel_index(i, self.transitions_depends.shape)
            for i in indices[::-1][:top_k]
        ]
        print('\nTop-%d unlikely transitions:' % (top_k))
        for i in range(top_k):
            print(
                '%d -> %d: %f'
                % (
                    bottom_trans[i][0],
                    bottom_trans[i][1],
                    self.transitions_depends[bottom_trans[i]],
                )
            )

    def print_features(self, top_k = 10):
        """
        Print important top-k features

        Parameters
        ----------
        top_k : int
        """
        if not isinstance(top_k, int):
            raise ValueError('input must be an integer')
        _features = self.features.sum(axis = 1)
        indices = np.argsort(_features)[::-1]
        rev_indices = indices[::-1]
        print('Top-%d positive:' % (top_k))
        for i in range(top_k):
            print(
                '%s: %f'
                % (
                    self._settings['idx2word'][str(indices[i])],
                    _features[indices[i]],
                )
            )

        print('\nTop-%d negative:' % (top_k))
        for i in range(top_k):
            print(
                '%s: %f'
                % (
                    self._settings['idx2word'][str(rev_indices[i])],
                    _features[rev_indices[i]],
                )
            )

    def predict(self, string):
        """
        Tagging a string.

        Parameters
        ----------
        string : str

        Returns
        -------
        string: tagged string
        """
        if not isinstance(string, str):
            raise ValueError('input must be a string')
        string = entities_textcleaning(string)
        if len(string) > 120:
            raise Exception(
                'Dependency parsing only able to accept string less than 120 words'
            )
        batch_x = char_str_idx([string], self._settings['word2idx'], 2)
        batch_x_char = generate_char_seq(
            [string], self._settings['char2idx'], 2
        )
        tagging, depend = self._sess.run(
            [self._logits, self._logits_depends],
            feed_dict = {self._X: batch_x, self._X_char: batch_x_char},
        )
        tagging = [self._settings['idx2tag'][i] for i in tagging[0]]
        depend = depend[0] - 1
        for i in range(len(depend)):
            if depend[i] == 0 and tagging[i] != 'root':
                tagging[i] = 'UNK'
            elif depend[i] != 0 and tagging[i] == 'root':
                tagging[i] = 'UNK'
            elif depend[i] > len(tagging):
                depend[i] = len(tagging)
        return (
            [(string[i], tagging[i]) for i in range(len(depend))],
            [(string[i], depend[i]) for i in range(len(depend))],
        )


class TAGGING:
    def __init__(
        self,
        X,
        X_char,
        logits,
        settings,
        sess,
        model,
        transitions,
        features,
        is_lower = True,
        story = None,
    ):
        self._X = X
        self._X_char = X_char
        self._logits = logits
        self._settings = settings
        self._sess = sess
        self._model = model
        self._is_lower = is_lower
        self._story = story
        self._settings['idx2tag'] = {
            int(k): v for k, v in self._settings['idx2tag'].items()
        }
        self.transitions, self.features = self._sess.run(
            [transitions, features]
        )

    def print_transitions(self, top_k = 10):
        """
        Print important top-k transitions

        Parameters
        ----------
        top_k : int
        """
        if not isinstance(top_k, int):
            raise ValueError('input must be an integer')
        print('Top-%d likely transitions:' % (top_k))
        indices = np.argsort(self.transitions.flatten())[::-1]
        top_trans = [
            np.unravel_index(i, self.transitions.shape) for i in indices[:top_k]
        ]
        for i in range(top_k):
            print(
                '%s -> %s: %f'
                % (
                    self._settings['idx2tag'][top_trans[i][0]],
                    self._settings['idx2tag'][top_trans[i][1]],
                    self.transitions[top_trans[i]],
                )
            )

        bottom_trans = [
            np.unravel_index(i, self.transitions.shape)
            for i in indices[::-1][:top_k]
        ]
        print('\nTop-%d unlikely transitions:' % (top_k))
        for i in range(top_k):
            print(
                '%s -> %s: %f'
                % (
                    self._settings['idx2tag'][bottom_trans[i][0]],
                    self._settings['idx2tag'][bottom_trans[i][1]],
                    self.transitions[bottom_trans[i]],
                )
            )

    def print_features(self, top_k = 10):
        """
        Print important top-k features

        Parameters
        ----------
        top_k : int
        """
        if not isinstance(top_k, int):
            raise ValueError('input must be an integer')
        _features = self.features.sum(axis = 1)
        indices = np.argsort(_features)[::-1]
        rev_indices = indices[::-1]
        print('Top-%d positive:' % (top_k))
        for i in range(top_k):
            print(
                '%s: %f'
                % (
                    self._settings['idx2word'][str(indices[i])],
                    _features[indices[i]],
                )
            )

        print('\nTop-%d negative:' % (top_k))
        for i in range(top_k):
            print(
                '%s: %f'
                % (
                    self._settings['idx2word'][str(rev_indices[i])],
                    _features[rev_indices[i]],
                )
            )

    def predict(self, string):
        """
        Tagging a string.

        Parameters
        ----------
        string : str

        Returns
        -------
        string: tagged string
        """
        if not isinstance(string, str):
            raise ValueError('input must be a string')
        string = string.lower() if self._is_lower else string
        string = entities_textcleaning(string)
        batch_x = char_str_idx([string], self._settings['word2idx'], 2)
        batch_x_char = generate_char_seq(
            [string], self._settings['char2idx'], 2
        )
        if self._model == 'entity-network':
            batch_x_expand = np.expand_dims(batch_x, axis = 1)
            predicted = self._sess.run(
                self._logits,
                feed_dict = {
                    self._X: batch_x,
                    self._X_char: batch_x_char,
                    self._story: batch_x_expand,
                },
            )[0]
        else:
            predicted = self._sess.run(
                self._logits,
                feed_dict = {self._X: batch_x, self._X_char: batch_x_char},
            )[0]
        return [
            (string[i], self._settings['idx2tag'][predicted[i]])
            for i in range(len(predicted))
        ]


class SOFTMAX:
    def __init__(
        self,
        X,
        logits,
        sess,
        mode,
        dictionary,
        alphas = None,
        input_mask = None,
        segment_ids = None,
        is_training = None,
        dropout_keep_prob = None,
        story = None,
        maxlen = 80,
        label = ['negative', 'positive'],
    ):
        self._X = X
        self._logits = logits
        self._sess = sess
        self._mode = mode
        self._dictionary = dictionary
        self._alphas = alphas
        self._input_mask = input_mask
        self._segment_ids = segment_ids
        self._is_training = is_training
        self._dropout_keep_prob = dropout_keep_prob
        self._story = story
        self._maxlen = maxlen
        self._label = label

    def get_dictionary(self):
        return self._dictionary

    def predict(self, string, get_proba = False):
        """
        classify a string.

        Parameters
        ----------
        string : str

        Returns
        -------
        dictionary: results
        """
        if not isinstance(string, str):
            raise ValueError('input must be a string')
        string = _classification_textcleaning_stemmer(string, attention = True)
        splitted = string[1].split()
        if self._mode in ['entity-network', 'bert']:
            batch_x = str_idx(
                [string[0]], self._dictionary, self._maxlen, UNK = 3
            )
        else:
            batch_x = str_idx(
                [string[0]], self._dictionary, len(splitted), UNK = 3
            )
        if self._mode in ['luong', 'bahdanau', 'hierarchical']:
            probs, alphas = self._sess.run(
                [tf.nn.softmax(self._logits), self._alphas],
                feed_dict = {self._X: batch_x},
            )
            if self._mode == 'hierarchical':
                alphas = alphas[0]
            words = []
            for i in range(alphas.shape[0]):
                words.append([splitted[i], alphas[i]])
        if self._mode in ['bidirectional', 'fast-text']:
            probs = self._sess.run(
                tf.nn.softmax(self._logits), feed_dict = {self._X: batch_x}
            )
        if self._mode == 'bert':
            np_mask = np.ones((1, self._maxlen), dtype = np.int32)
            np_segment = np.ones((1, self._maxlen), dtype = np.int32)
            probs = self._sess.run(
                tf.nn.softmax(self._logits),
                feed_dict = {
                    self._X: batch_x,
                    self._input_mask: np_mask,
                    self._segment_ids: np_segment,
                    self._is_training: False,
                },
            )
        if self._mode == 'entity-network':
            batch_x_expand = np.expand_dims(batch_x, axis = 1)
            probs = self._sess.run(
                tf.nn.softmax(self._logits),
                feed_dict = {
                    self._X: batch_x,
                    self._story: batch_x_expand,
                    self._dropout_keep_prob: 1.0,
                },
            )

        if get_proba:
            dict_result = {}
            for no, label in enumerate(self._label):
                dict_result[label] = probs[0, no]
            if self._mode in ['luong', 'bahdanau', 'hierarchical']:
                dict_result['attention'] = words
            return dict_result
        else:
            return self._label[np.argmax(probs[0])]

    def predict_batch(self, strings, get_proba = False):
        """
        classify list of strings

        Parameters
        ----------
        strings : list

        Returns
        -------
        list_dictionaries: list of results
        """
        if not isinstance(strings, list):
            raise ValueError('input must be a list')
        if not isinstance(strings[0], str):
            raise ValueError('input must be list of strings')
        strings = [_classification_textcleaning_stemmer(i) for i in strings]
        maxlen = max([len(i.split()) for i in strings])
        if self._mode in ['entity-network', 'bert']:
            batch_x = str_idx(strings, self._dictionary, self._maxlen, UNK = 3)
        else:
            batch_x = str_idx(strings, self._dictionary, maxlen, UNK = 3)

        if self._mode not in ['bert', 'entity-network']:
            probs = self._sess.run(
                tf.nn.softmax(self._logits), feed_dict = {self._X: batch_x}
            )
        if self._mode == 'bert':
            np_mask = np.ones((len(batch_x), self._maxlen), dtype = np.int32)
            np_segment = np.ones((len(batch_x), self._maxlen), dtype = np.int32)
            probs = self._sess.run(
                tf.nn.softmax(self._logits),
                feed_dict = {
                    self._X: batch_x,
                    self._input_mask: np_mask,
                    self._segment_ids: np_segment,
                    self._is_training: False,
                },
            )
        if self._mode == 'entity-network':
            batch_x_expand = np.expand_dims(batch_x, axis = 1)
            probs = self._sess.run(
                tf.nn.softmax(self._logits),
                feed_dict = {
                    self._X: batch_x,
                    self._story: batch_x_expand,
                    self._dropout_keep_prob: 1.0,
                },
            )

        results = []
        if get_proba:
            for prob in probs:
                dict_result = {}
                for no, label in enumerate(self._label):
                    dict_result[label] = prob[no]
                results.append(dict_result)
        else:
            probs = np.argmax(probs, 1)
            for prob in probs:
                results.append(self._label[prob])
        return results


class SIGMOID:
    def __init__(
        self,
        X,
        logits,
        sess,
        mode,
        dictionary,
        alphas = None,
        input_mask = None,
        segment_ids = None,
        is_training = None,
        dropout_keep_prob = None,
        story = None,
        maxlen = 80,
    ):
        self._X = X
        self._logits = logits
        self._sess = sess
        self._mode = mode
        self._dictionary = dictionary
        self._alphas = alphas
        self._input_mask = input_mask
        self._segment_ids = segment_ids
        self._is_training = is_training
        self._dropout_keep_prob = dropout_keep_prob
        self._story = story
        self._maxlen = maxlen
        self._label = [
            'toxic',
            'severe_toxic',
            'obscene',
            'threat',
            'insult',
            'identity_hate',
        ]

    def get_dictionary(self):
        return self._dictionary

    def predict(self, string, get_proba = False):
        """
        classify a string.

        Parameters
        ----------
        string : str

        Returns
        -------
        dictionary: results
        """
        if not isinstance(string, str):
            raise ValueError('input must be a string')
        string = _classification_textcleaning_stemmer(string, attention = True)
        splitted = string[1].split()
        if self._mode in ['entity-network', 'bert']:
            batch_x = str_idx(
                [string[0]], self._dictionary, self._maxlen, UNK = 3
            )
        else:
            batch_x = str_idx(
                [string[0]], self._dictionary, len(splitted), UNK = 3
            )
        if self._mode in ['luong', 'bahdanau', 'hierarchical']:
            probs, alphas = self._sess.run(
                [tf.nn.sigmoid(self._logits), self._alphas],
                feed_dict = {self._X: batch_x},
            )
            if self._mode == 'hierarchical':
                alphas = alphas[0]
            words = []
            for i in range(alphas.shape[0]):
                words.append([splitted[i], alphas[i]])
        if self._mode in ['fast-text']:
            probs = self._sess.run(
                tf.nn.softmax(self._logits), feed_dict = {self._X: batch_x}
            )
        if self._mode == 'entity-network':
            batch_x_expand = np.expand_dims(batch_x, axis = 1)
            probs = self._sess.run(
                tf.nn.softmax(self._logits),
                feed_dict = {
                    self._X: batch_x,
                    self._story: batch_x_expand,
                    self._dropout_keep_prob: 1.0,
                },
            )
        if get_proba:
            dict_result = {}
            for no, label in enumerate(self._label):
                dict_result[label] = probs[0, no]
            if self._mode in ['luong', 'bahdanau', 'hierarchical']:
                dict_result['attention'] = words
            return dict_result
        else:
            result = []
            probs = np.around(probs[0])
            for no, label in enumerate(self._label):
                if probs[no]:
                    result.append(label)
            return result

    def predict_batch(self, strings, get_proba = False):
        """
        classify list of strings

        Parameters
        ----------
        strings : list

        Returns
        -------
        list_dictionaries: list of results
        """
        if not isinstance(strings, list):
            raise ValueError('input must be a list')
        if not isinstance(strings[0], str):
            raise ValueError('input must be list of strings')
        strings = [
            _classification_textcleaning_stemmer(i, attention = True)[0]
            for i in strings
        ]
        maxlen = max([len(i.split()) for i in strings])
        if self._mode in ['entity-network']:
            batch_x = str_idx(strings, self._dictionary, self._maxlen, UNK = 3)
        else:
            batch_x = str_idx(strings, self._dictionary, maxlen, UNK = 3)
        if self._mode not in ['entity-network']:
            probs = self._sess.run(
                tf.nn.sigmoid(self._logits), feed_dict = {self._X: batch_x}
            )
        if self._mode == 'entity-network':
            batch_x_expand = np.expand_dims(batch_x, axis = 1)
            probs = self._sess.run(
                tf.nn.sigmoid(self._logits),
                feed_dict = {
                    self._X: batch_x,
                    self._story: batch_x_expand,
                    self._dropout_keep_prob: 1.0,
                },
            )
        results = []
        if get_proba:
            for prob in probs:
                dict_result = {}
                for no, label in enumerate(self._label):
                    dict_result[label] = prob[no]
                results.append(dict_result)
        else:
            probs = np.around(probs)
            for prob in probs:
                list_result = []
                for no, label in enumerate(self._label):
                    if prob[no]:
                        list_result.append(label)
                results.append(list_result)

        return results


class DEEP_LANG:
    def __init__(self, path, vectorizer, label):
        self._graph = tf.Graph()
        with self._graph.as_default():
            self._model = _LANG_MODEL()
            self._sess = tf.InteractiveSession()
            self._sess.run(tf.global_variables_initializer())
            saver = tf.train.Saver(tf.trainable_variables())
            saver.restore(self._sess, path + '/model.ckpt')
        self._vectorizer = vectorizer
        self._label = label

    def predict(self, string, get_proba = False):
        """
        classify a string.

        Parameters
        ----------
        string : str

        Returns
        -------
        dictionary: results
        """
        if not isinstance(string, str):
            raise ValueError('input must be a string')
        string = language_detection_textcleaning(string)
        transformed = self._vectorizer.transform([string])
        batch_x = _convert_sparse_matrix_to_sparse_tensor(transformed)
        probs = self._sess.run(
            tf.nn.softmax(self._model.logits),
            feed_dict = {self._model.X: batch_x[0], self._model.W: batch_x[1]},
        )[0]
        if get_proba:
            return {self._label[no]: i for no, i in enumerate(probs)}
        else:
            return self._label[np.argmax(probs)]

    def predict_batch(self, strings, get_proba = False):
        """
        classify list of strings

        Parameters
        ----------
        strings : list

        Returns
        -------
        list_dictionaries: list of results
        """
        if not isinstance(strings, list):
            raise ValueError('input must be a list')
        if not isinstance(strings[0], str):
            raise ValueError('input must be list of strings')
        strings = [language_detection_textcleaning(i) for i in strings]
        transformed = self._vectorizer.transform(strings)
        batch_x = _convert_sparse_matrix_to_sparse_tensor(transformed)
        probs = self._sess.run(
            tf.nn.softmax(self._model.logits),
            feed_dict = {self._model.X: batch_x[0], self._model.W: batch_x[1]},
        )
        dicts = []
        if get_proba:
            for i in range(probs.shape[0]):
                dicts.append(
                    {self._label[no]: k for no, k in enumerate(probs[i])}
                )
        else:
            probs = np.argmax(probs, 1)
            for prob in probs:
                dicts.append(self._label[prob])
        return dicts


class SPARSE_SOFTMAX:
    def __init__(
        self, path, vectorizer, label, output_size, embedded_size, vocab_size
    ):
        self._graph = tf.Graph()
        with self._graph.as_default():
            self._model = _SPARSE_SOFTMAX_MODEL(
                output_size, embedded_size, vocab_size
            )
            self._sess = tf.InteractiveSession()
            self._sess.run(tf.global_variables_initializer())
            saver = tf.train.Saver(tf.trainable_variables())
            saver.restore(self._sess, path + '/model.ckpt')
        self._vectorizer = vectorizer
        self._label = label

    def predict(self, string, get_proba = False):
        """
        classify a string.

        Parameters
        ----------
        string : str

        Returns
        -------
        dictionary: results
        """
        if not isinstance(string, str):
            raise ValueError('input must be a string')
        string = _classification_textcleaning_stemmer(string, attention = True)[
            0
        ]
        transformed = self._vectorizer.transform([string])
        batch_x = _convert_sparse_matrix_to_sparse_tensor(
            transformed, got_limit = False
        )
        probs = self._sess.run(
            tf.nn.softmax(self._model.logits),
            feed_dict = {self._model.X: batch_x[0], self._model.W: batch_x[1]},
        )[0]
        if get_proba:
            return {self._label[no]: i for no, i in enumerate(probs)}
        else:
            return self._label[np.argmax(probs)]

    def predict_batch(self, strings, get_proba = False):
        """
        classify list of strings

        Parameters
        ----------
        strings : list

        Returns
        -------
        list_dictionaries: list of results
        """
        if not isinstance(strings, list):
            raise ValueError('input must be a list')
        if not isinstance(strings[0], str):
            raise ValueError('input must be list of strings')
        strings = [
            _classification_textcleaning_stemmer(i, attention = True)[0]
            for i in strings
        ]
        transformed = self._vectorizer.transform(strings)
        batch_x = _convert_sparse_matrix_to_sparse_tensor(
            transformed, got_limit = False
        )
        probs = self._sess.run(
            tf.nn.softmax(self._model.logits),
            feed_dict = {self._model.X: batch_x[0], self._model.W: batch_x[1]},
        )
        dicts = []
        if get_proba:
            for i in range(probs.shape[0]):
                dicts.append(
                    {self._label[no]: k for no, k in enumerate(probs[i])}
                )
        else:
            probs = np.argmax(probs, 1)
            for prob in probs:
                dicts.append(self._label[prob])
        return dicts



influencer = {
    'mahathir': [
        'tun mahathir',
        'madey',
        'dr mahathir',
        'tun m',
        'mahathir',
        'madir',
        'dr m',
        'mahathir muhamad',
    ],
    'anwar ibrahim': ['anwar ibrahim', 'anwar'],
    'najib razak': [
        'najib razak',
        'ajib',
        'pakjib',
        'pok jib',
        'pak jib',
        'najib',
        'tan sri najib',
    ],
    'pakatan harapan': ['harapan', 'pakatan harapan', 'pakatan'],
    'syed saddiq': ['syed saddiq', 'syedsaddiq', 'syed', 'saddiq'],
    'parti keadilan rakyat': [
        'parti keadilan rakyat',
        'people justice party',
        'pkr',
    ],
    'umno': ['united malays', 'umno', 'united malays national organization'],
    'barisan nasional': ['bn', 'barisan nasional'],
    'parti islam semalaysia': [
        'parti islam',
        'malaysian islamic party',
        'parti islam semalaysia',
        'parti islam malaysia',
    ],
    'nurul izzah': ['nurul izzah', 'izzah'],
    'Tunku Ismail Idris': ['tunku ismail', 'tunku ismail idris', 'tmj'],
    'mca': ['malaysian chinese association', 'mca'],
    'democratic action party': ['democratic action party'],
    'parti amanah': [
        'national trust party',
        'amanah',
        'parti amanah',
        'parti amanah negara',
    ],
    'ppbm': [
        'parti pribumi bersatu malaysia',
        'ppbm',
        'malaysian united',
        'parti pribumi',
        'malaysian united indigenous party',
    ],
    'mic': [
        'malaysian indian',
        'malaysian indian congress',
        'indian congress',
        'mic',
    ],
    'Tun Daim Zainuddin': [
        'daim zainuddin',
        'daim',
        'tun daim',
        'tun daim zainuddin',
    ],
    'datuk seri abdul hadi awang': [
        'hadi',
        'datuk seri abdul hadi awang',
        'hadi awang',
    ],
    'majlis pakatan harapan': ['majlis pakatan harapan'],
    'wan azizah': ['kak wan', 'azizah', 'wan azizah'],
    'Parti Pribumi Bersatu Malaysia': [
        'parti pribumi bersatu malaysia',
        'ppbm',
        'parti bersatu',
        'parti pribumi',
    ],
    'Datuk Seri Azmin Ali': ['azmin ali', 'datuk seri azmin ali', 'azmin'],
    'Datuk Johari Abdul': ['johari', 'johari abdul', 'datuk johari abdul'],
    'Tengku Razaleigh Hamzah': ['tengku razaleigh hamzah', 'razaleigh'],
    'Tan Sri Dr Rais Yatim': ['rais yatim', 'tan sri dr rais yatim'],
    'rafizi ramli': ['rafizi', 'rafizi ramli'],
    'bersatu': ['bersatu'],
    'bernama': ['bernama'],
    'donald trump': ['trump', 'donald', 'donald trump'],
    'perkasa': ['pertubuhan pribumi perkasa', 'pertubuhan pribumi', 'perkasa'],
    'Tan Sri Mokhzani Mahathir': [
        'tan sri mokhzani mahathir',
        'mokhzani',
        'mokhzani mahathir',
    ],
    'Rais Yatim': ['rais yatim', 'rais'],
    'Anthony Loke Siew Fook': ['anthony loke siew fook', 'loke siew'],
    'rosmah mansur': ['rosmah', 'rosmah mansur'],
    'arul kanda': ['arul kanda', 'kanda'],
    'manchester united': ['manchester', 'manchester united'],
    'arsenal fc': ['arsenal fc', 'arsenal'],
    'liverpool fc': ['liverpool fc', 'liverpool'],
    'chelsea fc': ['chelsea', 'chelsea fc'],
    'manchester city': ['manchester city'],
    'fc barcelona': ['barcelona', 'fc barcelona'],
    'real madrid cf': ['madrid', 'real madrid', 'real madrid cf'],
    'fc bayern munich': ['bayern munich', 'fc bayern munich'],
    'juventus fc': ['juventus fc', 'juventus'],
    'zeti aziz': ['zeti aziz'],
    'robert kuok': ['robert kuok'],
    'hassan merican': ['hassan merican'],
    'ks jomo': ['ks jomo'],
    'jho low': ['jho low'],
    'kadir jasin': ['kadir jasin'],
    'zakir naik': ['zakir naik'],
    'bung mokhtar': ['bung mokhtar'],
    'shafie apdal': ['shafie apdal'],
    'ariff md yusof': ['ariff yusof', 'datuk ariff md yusof', 'ariff md yusof'],
    'felda': ['felda'],
    'dato vida': ['vida', 'dato vida'],
    'Jabatan Perancangan Bandar dan Desa': [
        'jabatan perancangan bandar dan desa'
    ],
}



import sys
import warnings

if not sys.warnoptions:
    warnings.simplefilter('ignore')

import re
import os
import numpy as np
import itertools
import collections
from unidecode import unidecode
from .._utils._utils import download_file
from ._tatabahasa import stopword_tatabahasa, stopwords, stopwords_calon
from ._english_words import _english_words
from .. import home
import json

STOPWORDS = set(stopwords + stopword_tatabahasa + stopwords_calon)
STOPWORD_CALON = set(stopwords_calon)
VOWELS = 'aeiou'
PHONES = ['sh', 'ch', 'ph', 'sz', 'cz', 'sch', 'rz', 'dz']
ENGLISH_WORDS = _english_words


def _isWord(word):
    if word:
        consecutiveVowels = 0
        consecutiveConsonents = 0
        for idx, letter in enumerate(word.lower()):
            vowel = True if letter in VOWELS else False
            if idx:
                prev = word[idx - 1]
                prevVowel = True if prev in VOWELS else False
                if not vowel and letter == 'y' and not prevVowel:
                    vowel = True
                if prevVowel != vowel:
                    consecutiveVowels = 0
                    consecutiveConsonents = 0
            if vowel:
                consecutiveVowels += 1
            else:
                consecutiveConsonents += 1
            if consecutiveVowels >= 3 or consecutiveConsonents > 3:
                return False
            if consecutiveConsonents == 3:
                subStr = word[idx - 2 : idx + 1]
                if any(phone in subStr for phone in PHONES):
                    consecutiveConsonents -= 1
                    continue
                return False
    return True


_list_laughing = {
    'huhu',
    'haha',
    'gaga',
    'hihi',
    'wkawka',
    'wkwk',
    'kiki',
    'keke',
    'huehue',
}


def malaya_textcleaning(string):
    """
    use by normalizer, spell
    remove links, hashtags, alias
    only accept A-Z, a-z
    remove any laugh
    remove any repeated char more than 2 times
    remove most of nonsense words
    """
    string = re.sub(
        'http\S+|www.\S+',
        '',
        ' '.join(
            [
                word
                for word in string.split()
                if word.find('#') < 0 and word.find('@') < 0
            ]
        ),
    )
    string = unidecode(string).replace('.', '. ').replace(',', ' , ')
    string = re.sub('[^\'"A-Za-z\- ]+', ' ', string)
    string = re.sub(r'[ ]+', ' ', string.lower()).strip()
    string = [word for word in string.lower().split() if _isWord(word)]
    string = [
        word
        for word in string
        if not any([laugh in word for laugh in _list_laughing])
        and word[: len(word) // 2] != word[len(word) // 2 :]
    ]
    string = ' '.join(string)
    string = (
        ''.join(''.join(s)[:2] for _, s in itertools.groupby(string))
    ).split()
    return ' '.join([word for word in string if word not in STOPWORDS])


def normalizer_textcleaning(string):
    """
    use by normalizer, spell
    remove links, hashtags, alias
    only accept A-Z, a-z
    remove any laugh
    remove any repeated char more than 2 times
    """
    string = re.sub(
        'http\S+|www.\S+',
        '',
        ' '.join(
            [
                word
                for word in string.split()
                if word.find('#') < 0 and word.find('@') < 0
            ]
        ),
    )
    string = re.sub('[^A-Za-z ]+', ' ', string)
    string = re.sub(r'[ ]+', ' ', string).strip()
    string = [
        word.title() if word.isupper() else word
        for word in string.split()
        if len(word)
    ]
    string = [
        word
        for word in string
        if not any([laugh in word for laugh in _list_laughing])
    ]
    string = ' '.join(string)
    return ''.join(''.join(s)[:2] for _, s in itertools.groupby(string))


def simple_textcleaning(string, lowering = True):
    """
    use by topic modelling
    only accept A-Z, a-z
    """
    string = unidecode(string)
    string = re.sub('[^A-Za-z ]+', ' ', string)
    return re.sub(r'[ ]+', ' ', string.lower() if lowering else string).strip()


def entities_textcleaning(string):
    """
    use by xgb entities, multinomial entities,
    xgb pos, xgb entities, char model, word model, concat model
    """
    string = re.sub('[^A-Za-z0-9\-\/ ]+', ' ', string)
    string = re.sub(r'[ ]+', ' ', string).strip()
    return [
        word.title() if word.isupper() else word
        for word in string.split()
        if len(word)
    ]


def summary_textcleaning(string):
    string = re.sub('[^A-Za-z0-9\-\/\'"\.\, ]+', ' ', string)
    return re.sub(r'[ ]+', ' ', string.lower()).strip()


def get_hashtags(string):
    return [hash.lower() for hash in re.findall('#(\w+)', string)]


def split_by_dot(string):
    string = re.sub(
        r'(?<!\d)\.(?!\d)',
        'SPLITTT',
        string.replace('\n', '').replace('/', ' '),
    )
    string = string.split('SPLITTT')
    return [re.sub(r'[ ]+', ' ', sentence).strip() for sentence in string]


def language_detection_textcleaning(string):
    string = re.sub(
        u'[0-9!@#$%^&*()_\-+{}|\~`\'";:?/.>,<]',
        ' ',
        string.lower(),
        flags = re.UNICODE,
    )
    return re.sub(r'[ ]+', ' ', string).strip()


def pos_entities_textcleaning(string):
    """
    use by text entities and pos
    remove links, hashtags, alias
    """
    string = re.sub(
        'http\S+|www.\S+',
        '',
        ' '.join(
            [i for i in string.split() if i.find('#') < 0 and i.find('@') < 0]
        ),
    )
    string = unidecode(string).replace('.', ' . ').replace(',', ' , ')
    string = re.sub('[^A-Za-z\- ]+', ' ', string)
    string = re.sub(r'[ ]+', ' ', string).strip()
    return ' '.join(
        [
            word.title() if word.isupper() else word
            for word in string.split()
            if len(word)
        ]
    )


def classification_textcleaning(string, no_stopwords = False, lowering = True):
    """
    stemmer, summarization, topic-modelling
    remove links, hashtags, alias
    """
    string = re.sub(
        'http\S+|www.\S+',
        '',
        ' '.join(
            [i for i in string.split() if i.find('#') < 0 and i.find('@') < 0]
        ),
    )
    string = unidecode(string).replace('.', ' . ').replace(',', ' , ')
    string = re.sub('[^A-Za-z ]+', ' ', string)
    string = re.sub(r'[ ]+', ' ', string).strip()
    if no_stopwords:
        string = ' '.join(
            [
                i
                for i in re.findall('[\\w\']+|[;:\-\(\)&.,!?"]', string)
                if len(i)
            ]
        )
    else:
        string = ' '.join(
            [
                i
                for i in re.findall('[\\w\']+|[;:\-\(\)&.,!?"]', string)
                if len(i) and i not in STOPWORDS
            ]
        )
    if lowering:
        return string.lower()
    else:
        return ' '.join(
            [
                word.title() if word.isupper() else word
                for word in string.split()
                if len(word)
            ]
        )


def separate_dataset(trainset):
    datastring = []
    datatarget = []
    for i in range(len(trainset.data)):
        data_ = trainset.data[i].split('\n')
        data_ = list(filter(None, data_))
        datastring += data_
        for n in range(len(data_)):
            datatarget.append(trainset.target[i])
    return datastring, datatarget


def print_topics_modelling(
    topics, feature_names, sorting, n_words = 20, return_df = True
):
    if return_df:
        try:
            import pandas as pd
        except:
            raise Exception(
                'pandas not installed. Please install it and try again.'
            )
    df = {}
    for i in range(topics):
        words = []
        for k in range(n_words):
            words.append(feature_names[sorting[i, k]])
        df['topic %d' % (i)] = words
    if return_df:
        return pd.DataFrame.from_dict(df)
    else:
        return df


def str_idx(corpus, dic, maxlen, UNK = 0):
    X = np.zeros((len(corpus), maxlen))
    for i in range(len(corpus)):
        for no, k in enumerate(corpus[i].split()[:maxlen][::-1]):
            X[i, -1 - no] = dic.get(k, UNK)
    return X


def stemmer_str_idx(corpus, dic, UNK = 3):
    X = []
    for i in corpus:
        ints = []
        for k in i:
            ints.append(dic.get(k, UNK))
        X.append(ints)
    return X


def pad_sentence_batch(sentence_batch, pad_int):
    padded_seqs = []
    seq_lens = []
    max_sentence_len = max([len(sentence) for sentence in sentence_batch])
    for sentence in sentence_batch:
        padded_seqs.append(
            sentence + [pad_int] * (max_sentence_len - len(sentence))
        )
        seq_lens.append(len(sentence))
    return padded_seqs, seq_lens


def char_str_idx(corpus, dic, UNK = 2):
    maxlen = max([len(i) for i in corpus])
    X = np.zeros((len(corpus), maxlen))
    for i in range(len(corpus)):
        for no, k in enumerate(corpus[i][:maxlen]):
            X[i, no] = dic.get(k, UNK)
    return X


def generate_char_seq(batch, dic, UNK = 2):
    maxlen_c = max([len(k) for k in batch])
    x = [[len(i) for i in k] for k in batch]
    maxlen = max([j for i in x for j in i])
    temp = np.zeros((len(batch), maxlen_c, maxlen), dtype = np.int32)
    for i in range(len(batch)):
        for k in range(len(batch[i])):
            for no, c in enumerate(batch[i][k][::-1]):
                temp[i, k, -1 - no] = dic.get(c, UNK)
    return temp


def build_dataset(words, n_words, included_prefix = True):
    count = (
        [['GO', 0], ['PAD', 1], ['EOS', 2], ['UNK', 3]]
        if included_prefix
        else []
    )
    count.extend(collections.Counter(words).most_common(n_words))
    dictionary = dict()
    for word, _ in count:
        dictionary[word] = len(dictionary)
    data = list()
    for word in words:
        index = dictionary.get(word, 3)
        data.append(index)
    reversed_dictionary = dict(zip(dictionary.values(), dictionary.keys()))
    return data, count, dictionary, reversed_dictionary



import os, shutil
from pathlib import Path
import sys

try:
    os.mkdir(str(Path.home())+'/Malaya')
except:
    pass
    
with open(str(Path.home())+'/Malaya/version', 'w') as fopen:
    fopen.write('0.0')
import malaya

def test_make_directory():
    del sys.modules['malaya']
    shutil.rmtree(str(Path.home())+'/Malaya')
    import malaya
    with open(str(Path.home())+'/Malaya/version', 'w') as fopen:
        fopen.write('0.0')
    del sys.modules['malaya']
    import malaya

def test_directory():
    with open(str(Path.home())+'/Malaya/something', 'w') as fopen:
        fopen.write('something')
    del sys.modules['malaya']
    import malaya
    assert os.path.exists(str(Path.home())+'/Malaya')

def test_print():
    malaya.describe_pos_malaya()
    malaya.describe_pos()
    malaya.describe_entities_malaya()
    malaya.describe_entities()



import sys
import warnings

if not sys.warnoptions:
    warnings.simplefilter('ignore')

import tensorflow as tf
import numpy as np
from tqdm import tqdm
import re
import collections
import json
import os
from sklearn.metrics import pairwise_distances_argmin_min
from sklearn.cluster import KMeans
from .._utils._paths import PATH_SUMMARIZE, S3_PATH_SUMMARIZE
from ..texts._text_functions import split_by_dot, summary_textcleaning
from .._utils._utils import download_file, load_graph


def batch_sequence(sentences, dictionary, maxlen = 50):
    np_array = np.zeros((len(sentences), maxlen), dtype = np.int32)
    for no_sentence, sentence in enumerate(sentences):
        current_no = 0
        for no, word in enumerate(sentence.split()[: maxlen - 2]):
            np_array[no_sentence, no] = dictionary.get(word, 1)
            current_no = no
        np_array[no_sentence, current_no + 1] = 3
    return np_array


class DEEP_SUMMARIZER:
    def __init__(
        self, sess, x, logits, attention, dictionary, maxlen, model = None
    ):
        self._sess = sess
        self._X = x
        self._logits = logits
        self._attention = attention
        self.dictionary = dictionary
        self._maxlen = maxlen
        self._rev_dictionary = {v: k for k, v in self.dictionary.items()}
        self._model = model

    def summarize(self, corpus, top_k = 3, important_words = 3):
        """
        Summarize list of strings / corpus

        Parameters
        ----------
        corpus: str, list

        top_k: int, (default=3)
            number of summarized strings
        important_words: int, (default=3)
            number of important words

        Returns
        -------
        string: summarized string
        """
        if not isinstance(corpus, list):
            raise ValueError('corpus must be a list')
        if not isinstance(corpus[0], str):
            raise ValueError('corpus must be list of strings')
        if isinstance(corpus, str):
            corpus = split_by_dot(corpus)
        else:
            corpus = ' '.join(corpus)
            corpus = re.findall('(?=\S)[^.\n]+(?<=\S)', corpus)

        corpus = [summary_textcleaning(i) for i in corpus]
        sequences = batch_sequence(
            corpus, self.dictionary, maxlen = self._maxlen
        )
        encoded, attention = self._sess.run(
            [self._logits, self._attention],
            feed_dict = {self._X: np.array(sequences)},
        )
        attention = attention.sum(axis = 0)
        kmeans = KMeans(n_clusters = top_k, random_state = 0)
        kmeans = kmeans.fit(encoded)
        avg = []
        for j in range(top_k):
            idx = np.where(kmeans.labels_ == j)[0]
            avg.append(np.mean(idx))
        closest, _ = pairwise_distances_argmin_min(
            kmeans.cluster_centers_, encoded
        )
        indices = np.argsort(attention)[::-1]
        top_words = [self._rev_dictionary[i] for i in indices[:important_words]]
        ordering = sorted(range(top_k), key = lambda k: avg[k])
        return {
            'summary': '. '.join([corpus[closest[idx]] for idx in ordering]),
            'top-words': top_words,
        }


class Model:
    def __init__(
        self,
        vocabulary_size,
        maxlen = 50,
        learning_rate = 1e-3,
        embedding_size = 256,
        **kwargs,
    ):
        word_embeddings = tf.Variable(
            tf.random_uniform(
                [vocabulary_size, embedding_size], -np.sqrt(3), np.sqrt(3)
            )
        )
        self.output_size = embedding_size
        self.maxlen = maxlen
        self.embeddings = word_embeddings
        self.output_layer = tf.layers.Dense(vocabulary_size)
        self.output_layer.build(self.output_size)

        self.BEFORE = tf.placeholder(tf.int32, [None, maxlen])
        self.INPUT = tf.placeholder(tf.int32, [None, maxlen])
        self.AFTER = tf.placeholder(tf.int32, [None, maxlen])
        self.batch_size = tf.shape(self.INPUT)[0]

        self.get_thought = self.thought(self.INPUT)
        self.attention = tf.matmul(
            self.get_thought, tf.transpose(self.embeddings), name = 'attention'
        )
        self.fw_logits = self.decoder(self.get_thought, self.AFTER)
        self.bw_logits = self.decoder(self.get_thought, self.BEFORE)
        self.loss = self.calculate_loss(
            self.fw_logits, self.AFTER
        ) + self.calculate_loss(self.bw_logits, self.BEFORE)
        self.optimizer = tf.train.AdamOptimizer(learning_rate).minimize(
            self.loss
        )

    def get_embedding(self, inputs):
        return tf.nn.embedding_lookup(self.embeddings, inputs)

    def thought(self, inputs):
        encoder_in = self.get_embedding(inputs)
        fw_cell = tf.nn.rnn_cell.GRUCell(self.output_size)
        bw_cell = tf.nn.rnn_cell.GRUCell(self.output_size)
        sequence_length = tf.reduce_sum(tf.sign(inputs), axis = 1)
        with tf.variable_scope('thought_scope', reuse = False):
            rnn_output = tf.nn.bidirectional_dynamic_rnn(
                fw_cell,
                bw_cell,
                encoder_in,
                sequence_length = sequence_length,
                dtype = tf.float32,
            )[1]
            return sum(rnn_output)

    def decoder(self, thought, labels):
        main = tf.strided_slice(labels, [0, 0], [self.batch_size, -1], [1, 1])
        shifted_labels = tf.concat([tf.fill([self.batch_size, 1], 2), main], 1)
        decoder_in = self.get_embedding(shifted_labels)
        cell = tf.nn.rnn_cell.GRUCell(self.output_size)
        max_seq_lengths = tf.fill([self.batch_size], self.maxlen)
        helper = tf.contrib.seq2seq.TrainingHelper(
            decoder_in, max_seq_lengths, time_major = False
        )
        decoder = tf.contrib.seq2seq.BasicDecoder(cell, helper, thought)
        decoder_out = tf.contrib.seq2seq.dynamic_decode(decoder)[0].rnn_output
        return decoder_out

    def calculate_loss(self, outputs, labels):
        mask = tf.cast(tf.sign(labels), tf.float32)
        logits = self.output_layer(outputs)
        return tf.contrib.seq2seq.sequence_loss(logits, labels, mask)


def counter_words(sentences):
    word_counter = collections.Counter()
    word_list = []
    num_lines, num_words = (0, 0)
    for i in sentences:
        words = re.findall('[\\w\']+|[;:\-\(\)&.,!?"]', i)
        word_counter.update(words)
        word_list.extend(words)
        num_lines += 1
        num_words += len(words)
    return word_counter, word_list, num_lines, num_words


def build_dict(word_counter, vocab_size = 50000):
    count = [['PAD', 0], ['UNK', 1], ['START', 2], ['END', 3]]
    count.extend(word_counter.most_common(vocab_size))
    dictionary = dict()
    for word, _ in count:
        dictionary[word] = len(dictionary)
    return dictionary, {word: idx for idx, word in dictionary.items()}


def news_load_model():
    if not os.path.isfile(PATH_SUMMARIZE['news']['model']):
        print('downloading SUMMARIZE news frozen model')
        download_file(
            S3_PATH_SUMMARIZE['news']['model'], PATH_SUMMARIZE['news']['model']
        )
    if not os.path.isfile(PATH_SUMMARIZE['news']['setting']):
        print('downloading SUMMARIZE news dictionary')
        download_file(
            S3_PATH_SUMMARIZE['news']['setting'],
            PATH_SUMMARIZE['news']['setting'],
        )
    g = load_graph(PATH_SUMMARIZE['news']['model'])
    x = g.get_tensor_by_name('import/Placeholder_1:0')
    logits = g.get_tensor_by_name('import/thought_scope/add_1:0')
    attention = g.get_tensor_by_name('import/attention:0')
    sess = tf.InteractiveSession(graph = g)
    with open(PATH_SUMMARIZE['news']['setting']) as fopen:
        dictionary = json.load(fopen)
    return DEEP_SUMMARIZER(sess, x, logits, attention, dictionary, 100)


def wiki_load_model():
    if not os.path.isfile(PATH_SUMMARIZE['wiki']['model']):
        print('downloading SUMMARIZE wikipedia frozen model')
        download_file(
            S3_PATH_SUMMARIZE['wiki']['model'], PATH_SUMMARIZE['wiki']['model']
        )
    if not os.path.isfile(PATH_SUMMARIZE['wiki']['setting']):
        print('downloading SUMMARIZE wikipedia dictionary')
        download_file(
            S3_PATH_SUMMARIZE['wiki']['setting'],
            PATH_SUMMARIZE['wiki']['setting'],
        )
    g = load_graph(PATH_SUMMARIZE['wiki']['model'])
    x = g.get_tensor_by_name('import/Placeholder_1:0')
    logits = g.get_tensor_by_name('import/logits:0')
    attention = g.get_tensor_by_name('import/attention:0')
    sess = tf.InteractiveSession(graph = g)
    with open(PATH_SUMMARIZE['wiki']['setting']) as fopen:
        dictionary = json.load(fopen)
    return DEEP_SUMMARIZER(sess, x, logits, attention, dictionary, 50)


def train_model(
    train_X,
    train_Y_before,
    train_Y_after,
    epoch = 10,
    batch_size = 16,
    embedding_size = 256,
    maxlen = 100,
    vocab_size = 50000,
    **kwargs,
):
    if not vocab_size:
        vocab_size = len(set(filter(None, (' '.join(train_X)).split()))) + 1
    word_counter, _, _, _ = counter_words(train_X)
    dictionary, _ = build_dict(word_counter, vocab_size = vocab_size)
    _graph = tf.Graph()
    with _graph.as_default():
        model = Model(
            len(dictionary),
            embedding_size = embedding_size,
            batch_size = batch_size,
            maxlen = maxlen,
            **kwargs,
        )
        sess = tf.InteractiveSession()
        sess.run(tf.global_variables_initializer())
        saver = tf.train.Saver(tf.trainable_variables())

    for e in range(epoch):
        pbar = tqdm(range(0, len(train_X), batch_size), desc = 'minibatch loop')
        for i in pbar:
            batch_x = batch_sequence(
                train_X[i : min(i + batch_size, len(train_X))],
                dictionary,
                maxlen = maxlen,
            )
            batch_y_before = batch_sequence(
                train_Y_before[i : min(i + batch_size, len(train_X))],
                dictionary,
                maxlen = maxlen,
            )
            batch_y_after = batch_sequence(
                train_Y_after[i : min(i + batch_size, len(train_X))],
                dictionary,
                maxlen = maxlen,
            )
            loss, _ = sess.run(
                [model.loss, model.optimizer],
                feed_dict = {
                    model.BEFORE: batch_y_before,
                    model.INPUT: batch_x,
                    model.AFTER: batch_y_after,
                },
            )
            pbar.set_postfix(cost = loss)
    return sess, model, dictionary, saver


def load_skipthought(location, json):
    graph = tf.Graph()
    with graph.as_default():
        model = Model(
            len(json['dictionary']),
            embedding_size = json['embedding_size'],
            maxlen = json['maxlen'],
        )
        sess = tf.InteractiveSession()
        sess.run(tf.global_variables_initializer())
        saver = tf.train.Saver(tf.trainable_variables())
        saver.restore(sess, location + '/model.ckpt')
    return sess, model, saver



import sys
import warnings

if not sys.warnoptions:
    warnings.simplefilter('ignore')

import json
import pickle
import os
from ._utils._utils import (
    check_file,
    load_graph,
    check_available,
    generate_session,
)
from . import home
from ._utils._paths import PATH_TOXIC, S3_PATH_TOXIC
from ._models._sklearn_model import TOXIC
from ._models._tensorflow_model import SIGMOID


def available_deep_model():
    """
    List available deep learning toxicity analysis models.
    """
    return ['bahdanau', 'hierarchical', 'luong', 'fast-text', 'entity-network']


def multinomial(validate = True):
    """
    Load multinomial toxic model.

    Parameters
    ----------
    validate: bool, optional (default=True)
        if True, malaya will check model availability and download if not available.

    Returns
    -------
    TOXIC : malaya._models._sklearn_model.TOXIC class
    """
    if validate:
        check_file(PATH_TOXIC['multinomial'], S3_PATH_TOXIC['multinomial'])
    else:
        if not check_available(PATH_TOXIC['multinomial']):
            raise Exception(
                'toxic/multinomial is not available, please `validate = True`'
            )
    try:
        with open(PATH_TOXIC['multinomial']['model'], 'rb') as fopen:
            multinomial = pickle.load(fopen)
        with open(PATH_TOXIC['multinomial']['vector'], 'rb') as fopen:
            vectorize = pickle.load(fopen)
    except:
        raise Exception(
            "model corrupted due to some reasons, please run malaya.clear_cache('toxic/multinomial') and try again"
        )
    return TOXIC(multinomial, vectorize)


def logistic(validate = True):
    """
    Load logistic toxic model.

    Parameters
    ----------
    validate: bool, optional (default=True)
        if True, malaya will check model availability and download if not available.

    Returns
    -------
    TOXIC : malaya._models._sklearn_model.TOXIC class
    """
    if validate:
        check_file(PATH_TOXIC['logistic'], S3_PATH_TOXIC['logistic'])
    else:
        if not check_available(PATH_TOXIC['logistic']):
            raise Exception(
                'toxic/logistic is not available, please `validate = True`'
            )
    try:
        with open(PATH_TOXIC['logistic']['model'], 'rb') as fopen:
            logistic = pickle.load(fopen)
        with open(PATH_TOXIC['logistic']['vector'], 'rb') as fopen:
            vectorize = pickle.load(fopen)
    except:
        raise Exception(
            "model corrupted due to some reasons, please run malaya.clear_cache('toxic/logistic') and try again"
        )
    return TOXIC(logistic, vectorize)


def deep_model(model = 'luong', validate = True):
    """
    Load deep learning sentiment analysis model.

    Parameters
    ----------
    model : str, optional (default='luong')
        Model architecture supported. Allowed values:

        * ``'fast-text'`` - Fast-text architecture, embedded and logits layers only
        * ``'hierarchical'`` - LSTM with hierarchical attention architecture
        * ``'bahdanau'`` - LSTM with bahdanau attention architecture
        * ``'luong'`` - LSTM with luong attention architecture
        * ``'entity-network'`` - Recurrent Entity-Network architecture
    validate: bool, optional (default=True)
        if True, malaya will check model availability and download if not available.

    Returns
    -------
    TOXIC: malaya._models._tensorflow_model.SIGMOID class
    """
    if not isinstance(model, str):
        raise ValueError('model must be a string')
    model = model.lower()
    if model == 'fast-text':
        if validate:
            check_file(PATH_TOXIC['fast-text'], S3_PATH_TOXIC['fast-text'])
        else:
            if not check_available(PATH_TOXIC['fast-text']):
                raise Exception(
                    'toxic/fast-text is not available, please `validate = True`'
                )
        try:
            with open(PATH_TOXIC['fast-text']['setting'], 'r') as fopen:
                dictionary = json.load(fopen)['dictionary']
            g = load_graph(PATH_TOXIC['fast-text']['model'])
        except:
            raise Exception(
                "model corrupted due to some reasons, please run malaya.clear_cache('toxic/fast-text') and try again"
            )
        return SIGMOID(
            g.get_tensor_by_name('import/Placeholder:0'),
            g.get_tensor_by_name('import/logits:0'),
            generate_session(graph = g),
            model,
            dictionary,
        )
    elif model in ['bahdanau', 'luong', 'hierarchical']:
        if validate:
            check_file(PATH_TOXIC[model], S3_PATH_TOXIC[model])
        else:
            if not check_available(PATH_TOXIC[model]):
                raise Exception(
                    'toxic/%s is not available, please `validate = True`'
                    % (model)
                )
        try:
            with open(PATH_TOXIC[model]['setting'], 'r') as fopen:
                dictionary = json.load(fopen)['dictionary']
            g = load_graph(PATH_TOXIC[model]['model'])
        except:
            raise Exception(
                "model corrupted due to some reasons, please run malaya.clear_cache('toxic/%s') and try again"
                % (model)
            )
        return SIGMOID(
            g.get_tensor_by_name('import/Placeholder:0'),
            g.get_tensor_by_name('import/logits:0'),
            generate_session(graph = g),
            model,
            dictionary,
            alphas = g.get_tensor_by_name('import/alphas:0'),
        )
    elif model == 'entity-network':
        if validate:
            check_file(
                PATH_TOXIC['entity-network'], S3_PATH_TOXIC['entity-network']
            )
        else:
            if not check_available(PATH_TOXIC['entity-network']):
                raise Exception(
                    'toxic/entity-network is not available, please `validate = True`'
                )
        try:
            with open(PATH_TOXIC['entity-network']['setting'], 'r') as fopen:
                dictionary = json.load(fopen)
            g = load_graph(PATH_TOXIC['entity-network']['model'])
        except:
            raise Exception(
                "model corrupted due to some reasons, please run malaya.clear_cache('toxic/entity-network') and try again"
            )
        return SIGMOID(
            g.get_tensor_by_name('import/Placeholder_question:0'),
            g.get_tensor_by_name('import/logits:0'),
            generate_session(graph = g),
            model,
            dictionary,
            dropout_keep_prob = g.get_tensor_by_name(
                'import/Placeholder_dropout_keep_prob:0'
            ),
            story = g.get_tensor_by_name('import/Placeholder_story:0'),
        )
    else:
        raise Exception(
            'model sentiment not supported, please check supported models from malaya.toxic.available_deep_model()'
        )



import sys
import warnings

if not sys.warnoptions:
    warnings.simplefilter('ignore')

import numpy as np
import re
import random
from scipy.linalg import svd
from operator import itemgetter
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.utils import shuffle
from sklearn.decomposition import NMF, LatentDirichletAllocation
from .texts._text_functions import (
    summary_textcleaning,
    classification_textcleaning,
    STOPWORDS,
    split_by_dot,
)
from .stem import sastrawi
from ._models import _skip_thought
from .cluster import cluster_words


def deep_model_news():
    """
    Load skip-thought summarization deep learning model trained on news dataset.

    Returns
    -------
    DEEP_SUMMARIZER: malaya.skip_thought.DEEP_SUMMARIZER class
    """
    return _skip_thought.news_load_model()


def deep_model_wiki():
    """
    Load residual network with Bahdanau Attention summarization deep learning model trained on wikipedia dataset.

    Returns
    -------
    DEEP_SUMMARIZER: malaya.skip_thought.DEEP_SUMMARIZER class
    """
    print(
        'WARNING: this model is using convolutional based, Tensorflow-GPU above 1.10 may got a problem. Please downgrade to Tensorflow-GPU v1.8 if got any cuDNN error.'
    )
    return _skip_thought.wiki_load_model()


def train_skip_thought(
    corpus,
    epoch = 5,
    batch_size = 16,
    embedding_size = 256,
    maxlen = 50,
    vocab_size = None,
    stride = 1,
):
    """
    Train a deep skip-thought network for summarization agent

    Parameters
    ----------
    epoch: int, (default=5)
        iteration numbers
    batch_size: int, (default=32)
        batch size for every feed, batch size must <= size of corpus
    embedding_size: int, (default=256)
        vector size representation for a word
    maxlen: int, (default=50)
        max length of a string to be train
    vocab_size: int, (default=None)
        max vocabulary size, None for no limit
    stride: int, (default=1)
        stride size, skipping value for sentences

    Returns
    -------
    DEEP_SUMMARIZER: malaya.skip_thought.DEEP_SUMMARIZER class
    """
    assert isinstance(epoch, int), 'epoch must be an integer'
    assert isinstance(batch_size, int), 'batch_size must be an integer'
    assert isinstance(embedding_size, int), 'embedding_size must be an integer'
    assert isinstance(maxlen, int), 'maxlen must be an integer'
    assert vocab_size is None or isinstance(
        vocab_size, int
    ), 'vocab_size must be a None or an integer'
    assert (
        isinstance(corpus, list) and isinstance(corpus[0], str)
    ) or isinstance(corpus, str), 'input must be list of strings or a string'
    if isinstance(corpus, str):
        corpus = split_by_dot(corpus)
    else:
        corpus = ' '.join(corpus)
        corpus = re.findall('(?=\S)[^.\n]+(?<=\S)', corpus)
    corpus = [summary_textcleaning(i) for i in corpus]
    t_range = int((len(corpus) - 3) / stride + 1)
    left, middle, right = [], [], []
    for i in range(t_range):
        slices = corpus[i * stride : i * stride + 3]
        left.append(slices[0])
        middle.append(slices[1])
        right.append(slices[2])
    assert batch_size < len(left), 'batch size must smaller with corpus size'
    left, middle, right = shuffle(left, middle, right)
    sess, model, dictionary, _ = _skip_thought.train_model(
        middle,
        left,
        right,
        epoch = epoch,
        batch_size = batch_size,
        embedding_size = embedding_size,
        maxlen = maxlen,
        vocab_size = vocab_size,
    )
    return _skip_thought.DEEP_SUMMARIZER(
        sess,
        model.INPUT,
        model.get_thought,
        model.attention,
        dictionary,
        maxlen,
        model = model,
    )


def lsa(
    corpus,
    maintain_original = False,
    ngram = (1, 3),
    min_df = 2,
    top_k = 3,
    important_words = 3,
    return_cluster = True,
    **kwargs
):
    """
    summarize a list of strings using LSA.

    Parameters
    ----------
    corpus: list
    maintain_original: bool, (default=False)
        If False, will apply malaya.text_functions.classification_textcleaning
    ngram: tuple, (default=(1,3))
        n-grams size to train a corpus
    min_df: int, (default=2)
        minimum document frequency for a word
    top_k: int, (default=3)
        number of summarized strings
    important_words: int, (default=3)
        number of important words
    return_cluster: bool, (default=True)
        if True, will cluster important_words to similar texts

    Returns
    -------
    dictionary: result
    """
    assert isinstance(corpus, list) and isinstance(
        corpus[0], str
    ), 'input must be list of strings'
    assert isinstance(
        maintain_original, bool
    ), 'maintain_original must be a boolean'
    assert isinstance(top_k, int), 'top_k must be an integer'
    assert isinstance(
        important_words, int
    ), 'important_words must be an integer'
    assert isinstance(return_cluster, bool), 'return_cluster must be a boolean'
    if not isinstance(ngram, tuple):
        raise ValueError('ngram must be a tuple')
    if not len(ngram) == 2:
        raise ValueError('ngram size must equal to 2')

    corpus = [summary_textcleaning(i) for i in corpus]
    corpus = ' '.join(corpus)
    splitted_fullstop = re.findall('(?=\S)[^.\n]+(?<=\S)', corpus)
    splitted_fullstop = [
        classification_textcleaning(i) if not maintain_original else i
        for i in splitted_fullstop
        if len(i)
    ]
    stemmed = [sastrawi(i) for i in splitted_fullstop]
    tfidf = TfidfVectorizer(
        ngram_range = ngram, min_df = min_df, stop_words = STOPWORDS, **kwargs
    ).fit(stemmed)
    U, S, Vt = svd(tfidf.transform(stemmed).todense().T, full_matrices = False)
    summary = [
        (splitted_fullstop[i], np.linalg.norm(np.dot(np.diag(S), Vt[:, b]), 2))
        for i in range(len(splitted_fullstop))
        for b in range(len(Vt))
    ]
    summary = sorted(summary, key = itemgetter(1))
    summary = dict(
        (v[0], v) for v in sorted(summary, key = lambda summary: summary[1])
    ).values()
    summarized = '. '.join([a for a, b in summary][len(summary) - (top_k) :])
    indices = np.argsort(tfidf.idf_)[::-1]
    features = tfidf.get_feature_names()
    top_words = [features[i] for i in indices[:important_words]]
    if return_cluster:
        return {
            'summary': summarized,
            'top-words': top_words,
            'cluster-top-words': cluster_words(top_words),
        }
    return {'summary': summarized, 'top-words': top_words}


def nmf(
    corpus,
    maintain_original = False,
    ngram = (1, 3),
    min_df = 2,
    top_k = 3,
    important_words = 3,
    return_cluster = True,
    **kwargs
):
    """
    summarize a list of strings using NMF.

    Parameters
    ----------
    corpus: list
    maintain_original: bool, (default=False)
        If False, will apply malaya.text_functions.classification_textcleaning
    ngram: tuple, (default=(1,3))
        n-grams size to train a corpus
    top_k: int, (default=3)
        number of summarized strings
    important_words: int, (default=3)
        number of important words
    min_df: int, (default=2)
        minimum document frequency for a word
    return_cluster: bool, (default=True)
        if True, will cluster important_words to similar texts

    Returns
    -------
    dictionary: result
    """
    assert isinstance(corpus, list) and isinstance(
        corpus[0], str
    ), 'input must be list of strings'
    assert isinstance(
        maintain_original, bool
    ), 'maintain_original must be a boolean'
    assert isinstance(top_k, int), 'top_k must be an integer'
    assert isinstance(
        important_words, int
    ), 'important_words must be an integer'
    assert isinstance(return_cluster, bool), 'return_cluster must be a boolean'
    if not isinstance(ngram, tuple):
        raise ValueError('ngram must be a tuple')
    if not len(ngram) == 2:
        raise ValueError('ngram size must equal to 2')

    corpus = [summary_textcleaning(i) for i in corpus]
    corpus = ' '.join(corpus)
    splitted_fullstop = re.findall('(?=\S)[^.\n]+(?<=\S)', corpus)
    splitted_fullstop = [
        classification_textcleaning(i) if not maintain_original else i
        for i in splitted_fullstop
        if len(i)
    ]
    stemmed = [sastrawi(i) for i in splitted_fullstop]
    tfidf = TfidfVectorizer(
        ngram_range = ngram, min_df = min_df, stop_words = STOPWORDS, **kwargs
    ).fit(stemmed)
    densed_tfidf = tfidf.transform(stemmed).todense()
    nmf = NMF(len(splitted_fullstop)).fit(densed_tfidf)
    vectors = nmf.transform(densed_tfidf)
    components = nmf.components_.mean(axis = 1)
    summary = [
        (
            splitted_fullstop[i],
            np.linalg.norm(np.dot(np.diag(components), vectors[:, b]), 2),
        )
        for i in range(len(splitted_fullstop))
        for b in range(len(vectors))
    ]
    summary = sorted(summary, key = itemgetter(1))
    summary = dict(
        (v[0], v) for v in sorted(summary, key = lambda summary: summary[1])
    ).values()
    summarized = '. '.join([a for a, b in summary][len(summary) - (top_k) :])
    indices = np.argsort(tfidf.idf_)[::-1]
    features = tfidf.get_feature_names()
    top_words = [features[i] for i in indices[:important_words]]
    if return_cluster:
        return {
            'summary': summarized,
            'top-words': top_words,
            'cluster-top-words': cluster_words(top_words),
        }
    return {'summary': summarized, 'top-words': top_words}


def lda(
    corpus,
    maintain_original = False,
    ngram = (1, 3),
    min_df = 2,
    top_k = 3,
    important_words = 3,
    return_cluster = True,
    **kwargs
):
    """
    summarize a list of strings using LDA.

    Parameters
    ----------
    corpus: list
    maintain_original: bool, (default=False)
        If False, will apply malaya.text_functions.classification_textcleaning
    ngram: tuple, (default=(1,3))
        n-grams size to train a corpus
    min_df: int, (default=2)
        minimum document frequency for a word
    top_k: int, (default=3)
        number of summarized strings
    important_words: int, (default=3)
        number of important words
    return_cluster: bool, (default=True)
        if True, will cluster important_words to similar texts

    Returns
    -------
    dictionary: result
    """
    assert isinstance(corpus, list) and isinstance(
        corpus[0], str
    ), 'input must be list of strings'
    assert isinstance(
        maintain_original, bool
    ), 'maintain_original must be a boolean'
    assert isinstance(top_k, int), 'top_k must be an integer'
    assert isinstance(
        important_words, int
    ), 'important_words must be an integer'
    assert isinstance(return_cluster, bool), 'return_cluster must be a boolean'
    if not isinstance(ngram, tuple):
        raise ValueError('ngram must be a tuple')
    if not len(ngram) == 2:
        raise ValueError('ngram size must equal to 2')
    if not isinstance(min_df, int) or isinstance(min_df, float):
        raise ValueError('min_df must be an integer or a float')

    corpus = [summary_textcleaning(i) for i in corpus]
    corpus = ' '.join(corpus)
    splitted_fullstop = re.findall('(?=\S)[^.\n]+(?<=\S)', corpus)
    splitted_fullstop = [
        classification_textcleaning(i) if not maintain_original else i
        for i in splitted_fullstop
        if len(i)
    ]
    stemmed = [sastrawi(i) for i in splitted_fullstop]
    tfidf = TfidfVectorizer(
        ngram_range = ngram, min_df = min_df, stop_words = STOPWORDS, **kwargs
    ).fit(stemmed)
    densed_tfidf = tfidf.transform(stemmed).todense()
    lda = LatentDirichletAllocation(len(splitted_fullstop)).fit(densed_tfidf)
    vectors = lda.transform(densed_tfidf)
    components = lda.components_.mean(axis = 1)
    summary = [
        (
            splitted_fullstop[i],
            np.linalg.norm(np.dot(np.diag(components), vectors[:, b]), 2),
        )
        for i in range(len(splitted_fullstop))
        for b in range(len(vectors))
    ]
    summary = sorted(summary, key = itemgetter(1))
    summary = dict(
        (v[0], v) for v in sorted(summary, key = lambda summary: summary[1])
    ).values()
    summarized = '. '.join([a for a, b in summary][len(summary) - (top_k) :])
    indices = np.argsort(tfidf.idf_)[::-1]
    features = tfidf.get_feature_names()
    top_words = [features[i] for i in indices[:important_words]]
    if return_cluster:
        return {
            'summary': summarized,
            'top-words': top_words,
            'cluster-top-words': cluster_words(top_words),
        }
    return {'summary': summarized, 'top-words': top_words}



import sys
import warnings

if not sys.warnoptions:
    warnings.simplefilter('ignore')

import tensorflow as tf
from tqdm import tqdm
from ..texts._text_functions import str_idx, build_dataset


class Model:
    def __init__(
        self,
        size_layer = 128,
        num_layers = 1,
        embedded_size = 128,
        dict_size = 5000,
        learning_rate = 1e-3,
        output_size = 300,
        dropout = 0.8,
    ):
        def cells(size, reuse = False):
            cell = tf.nn.rnn_cell.LSTMCell(
                size, initializer = tf.orthogonal_initializer(), reuse = reuse
            )
            return tf.contrib.rnn.DropoutWrapper(
                cell, output_keep_prob = dropout
            )

        def birnn(inputs, scope):
            with tf.variable_scope(scope):
                for n in range(num_layers):
                    (out_fw, out_bw), (
                        state_fw,
                        state_bw,
                    ) = tf.nn.bidirectional_dynamic_rnn(
                        cell_fw = cells(size_layer // 2),
                        cell_bw = cells(size_layer // 2),
                        inputs = inputs,
                        dtype = tf.float32,
                        scope = 'bidirectional_rnn_%d' % (n),
                    )
                    inputs = tf.concat((out_fw, out_bw), 2)
                return tf.layers.dense(inputs[:, -1], output_size)

        self.X_left = tf.placeholder(tf.int32, [None, None])
        self.X_right = tf.placeholder(tf.int32, [None, None])
        self.Y = tf.placeholder(tf.float32, [None])
        self.batch_size = tf.shape(self.X_left)[0]
        encoder_embeddings = tf.Variable(
            tf.random_uniform([dict_size, embedded_size], -1, 1)
        )
        embedded_left = tf.nn.embedding_lookup(encoder_embeddings, self.X_left)
        embedded_right = tf.nn.embedding_lookup(
            encoder_embeddings, self.X_right
        )

        def contrastive_loss(y, d):
            tmp = y * tf.square(d)
            tmp2 = (1 - y) * tf.square(tf.maximum((1 - d), 0))
            return (
                tf.reduce_sum(tmp + tmp2)
                / tf.cast(self.batch_size, tf.float32)
                / 2
            )

        self.output_left = birnn(embedded_left, 'left')
        self.output_right = birnn(embedded_right, 'right')
        self.distance = tf.sqrt(
            tf.reduce_sum(
                tf.square(tf.subtract(self.output_left, self.output_right)),
                1,
                keepdims = True,
            )
        )
        self.distance = tf.div(
            self.distance,
            tf.add(
                tf.sqrt(
                    tf.reduce_sum(
                        tf.square(self.output_left), 1, keepdims = True
                    )
                ),
                tf.sqrt(
                    tf.reduce_sum(
                        tf.square(self.output_right), 1, keepdims = True
                    )
                ),
            ),
        )
        self.distance = tf.reshape(self.distance, [-1])
        self.cost = contrastive_loss(self.Y, self.distance)

        self.temp_sim = tf.subtract(
            tf.ones_like(self.distance), tf.rint(self.distance)
        )
        correct_predictions = tf.equal(self.temp_sim, self.Y)
        self.accuracy = tf.reduce_mean(tf.cast(correct_predictions, 'float'))
        self.optimizer = tf.train.AdamOptimizer(
            learning_rate = learning_rate
        ).minimize(self.cost)


def train_model(
    train_X_left,
    train_X_right,
    train_Y,
    epoch = 10,
    batch_size = 16,
    embedding_size = 256,
    output_size = 300,
    maxlen = 100,
    dropout = 0.8,
    num_layers = 1,
    **kwargs
):
    concat = (' '.join(train_X_left + train_X_right)).split()
    vocabulary_size = len(list(set(concat)))
    _, _, dictionary, reversed_dictionary = build_dataset(
        concat, vocabulary_size
    )
    _graph = tf.Graph()
    with _graph.as_default():
        sess = tf.InteractiveSession()
        model = Model(
            size_layer = embedding_size,
            num_layers = num_layers,
            embedded_size = embedding_size,
            dict_size = len(dictionary),
            output_size = output_size,
            dropout = dropout,
        )
        sess.run(tf.global_variables_initializer())
        saver = tf.train.Saver(tf.trainable_variables())

    vectors_left = str_idx(train_X_left, dictionary, maxlen, UNK = 3)
    vectors_right = str_idx(train_X_right, dictionary, maxlen, UNK = 3)
    for e in range(epoch):
        pbar = tqdm(
            range(0, len(vectors_left), batch_size), desc = 'minibatch loop'
        )
        for i in pbar:
            batch_x_left = vectors_left[
                i : min(i + batch_size, len(vectors_left))
            ]
            batch_x_right = vectors_right[
                i : min(i + batch_size, len(vectors_left))
            ]
            batch_y = train_Y[i : min(i + batch_size, len(vectors_left))]
            acc, loss, _ = sess.run(
                [model.accuracy, model.cost, model.optimizer],
                feed_dict = {
                    model.X_left: batch_x_left,
                    model.X_right: batch_x_right,
                    model.Y: batch_y,
                },
            )
            pbar.set_postfix(cost = loss, accuracy = acc)
    return sess, model, dictionary, saver, dropout


def load_siamese(location, json):
    graph = tf.Graph()
    with graph.as_default():
        model = Model(
            size_layer = json['embedding_size'],
            num_layers = json['num_layers'],
            embedded_size = json['embedding_size'],
            dict_size = len(json['dictionary']),
            output_size = json['output_size'],
            dropout = json['dropout'],
        )
        sess = tf.InteractiveSession()
        sess.run(tf.global_variables_initializer())
        saver = tf.train.Saver(tf.trainable_variables())
        saver.restore(sess, location + '/model.ckpt')
    return sess, model, saver
